from datetime import datetime, timezone, timedelta
from threading import Lock
from typing import Any, Tuple, Optional

from qottoauth.api.base import QottoAuthApi

__all__ = [
    'QottoAuthCachedApi',
]


class QueryInput:
    def __init__(
            self,
            name: str,
            variables: list[Tuple[str, str, Any]] = None, body: str = None,
    ):
        self.key = '||'.join((name, str(variables), str(body)))
        self.date = datetime.now(timezone.utc)
        self.result: Any = None


class QottoAuthCachedApi(QottoAuthApi):

    def __init__(
            self,
            qotto_auth_api: QottoAuthApi,
            cache_local_seconds: int = 60,
            cache_max_size: int = 100,
            cache_max_seconds: int = 3600,
    ):
        self._qotto_auth_api = qotto_auth_api

        self._cache_lock = Lock()
        self._cache: dict[str, QueryInput] = dict()

        self._cache_local_seconds = cache_local_seconds
        self._cache_max_size = cache_max_size
        self._cache_max_seconds = cache_max_seconds

        self._last_clean_cache = datetime.now(timezone.utc)

        self._cache_auth_metadata: Optional[dict] = None
        self._last_auth_metadata = datetime.now(timezone.utc)

    def _should_clean_cache(self, date: datetime) -> bool:
        with self._cache_lock:
            return self._last_clean_cache < date - timedelta(seconds=self._cache_local_seconds)

    def _clean_cache(self, since: datetime) -> None:
        with self._cache_lock:
            self._last_clean_cache = datetime.now(timezone.utc)
            for cached_input in list(self._cache.values()):
                if cached_input.date < since:
                    del self._cache[cached_input.key]

    def _get_cached(self, query_input: QueryInput) -> Any:
        with self._cache_lock:
            cached = self._cache.get(query_input.key)
            return cached.result if cached is not None else None

    def _set_cached(self, query_input: QueryInput, result: dict[str, Any]) -> None:
        with self._cache_lock:
            query_input.result = result
            self._cache[query_input.key] = query_input
            if len(self._cache) > self._cache_max_size:
                self._cache.pop(next(iter(self._cache)))

    def query(
            self,
            name: str,
            variables: list[Tuple[str, str, Any]] = None,
            body: str = None,
    ):
        # No cache for these queries
        if name in ('lastChangeDate'):
            return self._qotto_auth_api.query(name, variables, body)
        # The input
        input = QueryInput(name, variables, body)
        # Check if we need to update the cache
        if self._should_clean_cache(date=input.date):
            last_change_date = max(
                datetime.fromisoformat(self._qotto_auth_api.query('lastChangeDate')),
                input.date - timedelta(seconds=self._cache_max_seconds),
            )
            self._clean_cache(since=last_change_date)
        # Check if we have a cached result
        cached_result = self._get_cached(input)
        if cached_result is not None:
            return cached_result
        # Get the result from the API
        computed_result = self._qotto_auth_api.query(name=name, variables=variables, body=body)
        # Set the result in the cache
        self._set_cached(input, computed_result)
        # Return the result
        return computed_result

    def mutation(
            self, name: str, body: str,
            input_name: str = 'input', input_type: str = None, input_value: dict[str, Any] = None
    ):
        self._clean_cache(datetime.now(timezone.utc))
        return self._qotto_auth_api.mutation(name, body, input_name, input_type, input_value)

    def auth_metadata(self) -> dict[str, str]:
        if self._cache_auth_metadata is None or self._last_auth_metadata < datetime.now(timezone.utc) - timedelta(
                seconds=self._cache_local_seconds
        ):
            self._cache_auth_metadata = self._qotto_auth_api.auth_metadata()
        return self._cache_auth_metadata



import logging
from datetime import date, datetime
from typing import Any, Optional, Union

import jwt

from qottoauth import Matching, Namespace
from qottoauth.api import QottoAuthApi
from qottoauth.models import (Account, Actor, Application, Authorization,
                              CookiePair, Identity, Member, Organization,
                              Permission, Role, TokenInfo, User, UserRequest)

__all__ = [
    'QottoAuthService',
]

logger = logging.getLogger(__name__)


def _filter_variables(*variables: tuple[str, str, Any]) -> list[tuple[str, str, Any]]:
    return [variable for variable in variables if variable[2] is not None]


class QottoAuthService:

    def __init__(
            self,
            api: QottoAuthApi,
    ):
        self.api = api

    def application(self, id: str) -> Application:
        response = self.api.query(
            name='application',
            variables=_filter_variables(
                ('id', 'ID!', id),
            ),
            body=Application.body(),
        )
        return Application.from_dict(response)

    def applications(self, name_contains: str = None) -> list[Application]:
        response = self.api.query(
            name='applications',
            variables=_filter_variables(
                ('name_Icontains', 'String', name_contains),
            ),
            body='edges { node {' + Application.body() + '} }'
        )
        return [Application.from_dict(edge['node']) for edge in response['edges']]

    def create_application(self, code: str, name: str = None, description: str = None, uuid: str = None) -> Application:
        payload = self.api.mutation(
            name='createApplication',
            input_value=dict(uuid=uuid, code=code, name=name, description=description),
            body='application {' + Application.body() + '}',
        )
        return Application.from_dict(payload['application'])

    def register_application(self, code: str, name: str = None, description: str = None, uuid=None) -> Application:
        payload = self.api.mutation(
            name='registerApplication',
            input_value=dict(uuid=uuid, code=code, name=name, description=description),
            body='application { ' + Application.body() + ' }',
        )
        return Application.from_dict(payload['application'])

    def delete_application(self, application: Union[Application, str]) -> bool:
        application_id = application.id if isinstance(application, Application) else application
        response = self.api.mutation(
            name='deleteApplication',
            input_value=dict(applicationId=application_id),
            body='deleted',
        )
        return bool(response['deleted'])

    def permission(self, id: str) -> Permission:
        response = self.api.query(
            name='permission',
            variables=[
                ('id', 'ID!', id),
            ],
            body=Permission.body(),
        )
        return Permission.from_dict(response)

    def permissions(
            self,
            name_contains: str = None,
            application: Union[Application, str] = None,
            authorization: Union[Authorization, str] = None,
    ) -> list[Permission]:
        application_id = application.id if isinstance(application, Application) else application
        authorization_id = authorization.id if isinstance(authorization, Authorization) else authorization
        response = self.api.query(
            name='permissions',
            variables=_filter_variables(
                ('name_Icontains', 'String', name_contains),
                ('application_Id', 'String', application_id),
                ('authorization_Id', 'String', authorization_id),
            ),
            body='edges { node { ' + Permission.body() + ' } }',
        )
        return [Permission.from_dict(edge['node']) for edge in response['edges']]

    def create_permission(
            self, application: Union[Application, str],
            code: str, name: str = None, description: str = None, uuid: str = None,
    ) -> Permission:
        application_id = application.id if isinstance(application, Application) else application
        payload = self.api.mutation(
            name='createPermission',
            input_value=dict(applicationId=application_id, uuid=uuid, code=code, name=name, description=description),
            body='permission { ' + Permission.body() + ' }',
        )
        return Permission.from_dict(payload['permission'])

    def register_permission(
            self,
            application: Union[Application, str],
            code: str, name: str = None, description: str = None, uuid: str = None,
    ) -> Permission:
        application_id = application.id if isinstance(application, Application) else application
        payload = self.api.mutation(
            name='registerPermission',
            input_value=dict(applicationId=application_id, uuid=uuid, code=code, name=name, description=description),
            body='permission { ' + Permission.body() + ' }',
        )
        return Permission.from_dict(payload['permission'])

    def register_permissions(
            self,
            application: Union[Application, str],
            permissions: list[dict[str, Any]],
    ) -> list[Permission]:
        application_id = application.id if isinstance(application, Application) else application
        payload = self.api.mutation(
            name='registerPermissions',
            input_value=dict(applicationId=application_id, permissions=permissions),
            body='permissions { ' + Permission.body() + ' }',
        )
        return [Permission.from_dict(permission) for permission in payload['permissions']]

    def delete_permission(self, permission: Union[Permission, str]) -> bool:
        permission_id = permission.id if isinstance(permission, Permission) else permission
        payload = self.api.mutation(
            name='deletePermission',
            input_value=dict(permissionId=permission_id),
            body='deleted',
        )
        return bool(payload['deleted'])

    def organization(self, id: str) -> Organization:
        response = self.api.query(
            name='organization',
            variables=_filter_variables(
                ('id', 'ID!', id),
            ),
            body='id name namespace',
        )
        return Organization.from_dict(response)

    def organizations(
            self,
            name_contains: str = None, namespace_contains: str = None,
    ) -> list[Organization]:
        response = self.api.query(
            name='organizations',
            variables=_filter_variables(
                ('name_Icontains', 'String', name_contains),
                ('namespace_Icontains', 'String', namespace_contains),
            ),
            body='edges { node { id name namespace } }',
        )
        return [Organization.from_dict(edge['node']) for edge in response['edges']]

    def create_organization(
            self,
            namespace: Union[Namespace, str], name: str = None, uuid: str = None,
    ) -> Organization:
        payload = self.api.mutation(
            name='createOrganization',
            input_value=dict(
                uuid=uuid,
                name=name,
                namespace=str(namespace),
            ),
            body='organization { ' + Organization.body() + ' }',
        )
        return Organization.from_dict(payload['organization'])

    def register_organization(
            self, namespace: Union[Namespace, str], name: str = None, uuid: str = None,
    ) -> Organization:
        payload = self.api.mutation(
            name='registerOrganization',
            input_value=dict(
                uuid=uuid,
                name=name,
                namespace=str(namespace),
            ),
            body='organization { ' + Organization.body() + ' }',
        )
        return Organization.from_dict(payload['organization'])

    def delete_organization(self, organization: Union[Organization, str]) -> bool:
        organization_id = organization.id if isinstance(organization, Organization) else organization
        payload = self.api.mutation(
            name='deleteOrganization',
            input_value=dict(organizationId=organization_id),
            body='deleted',
        )
        return bool(payload['deleted'])

    def user(self, id: str) -> User:
        response = self.api.query(
            name='user',
            variables=_filter_variables(
                ('id', 'ID!', id),
            ),
            body=User.body(),
        )
        return User.from_dict(response)

    def users(
            self,
            name_contains: str = None,
            associated: bool = None
    ) -> list[User]:
        response = self.api.query(
            name='users',
            variables=_filter_variables(
                ('associated', 'Boolean', associated),
                ('name_Icontains', 'String', name_contains),
            ),
            body='edges { node { ' + User.body() + ' } }',
        )
        return [User.from_dict(edge['node']) for edge in response['edges']]

    def create_user(
            self,
            name: str,
            email: str,
            given_name: str = None,
            middle_name: str = None,
            family_name: str = None,
            picture: str = None,
            birthdate: Union[date, str] = None,
            country: str = None,
            address: str = None,
            phone_number: str = None,
            job_title: str = None,
            is_superuser: bool = False,
            uuid: str = None,
    ) -> User:
        birthdate = date.fromisoformat(birthdate) if isinstance(birthdate, str) else birthdate
        payload = self.api.mutation(
            name='createUser',
            input_value=dict(
                uuid=uuid,
                isSuperuser=is_superuser,
                name=name,
                email=email,
                givenName=given_name,
                middleName=middle_name,
                familyName=family_name,
                picture=picture,
                birthdate=birthdate.isoformat() if birthdate else None,
                country=country,
                address=address,
                phoneNumber=phone_number,
                jobTitle=job_title,
            ),
            body='user { ' + User.body() + ' }',
        )
        return User.from_dict(payload['user'])

    def update_user(
            self, user: Union[User, str],
            add_identity: Union[Identity, str] = None, remove_identity: Union[Identity, str] = None,
            set_name: str = None, set_email: str = None,
            set_given_name: str = None, set_middle_name: str = None, set_family_name: str = None,
            set_picture: str = None, set_birthdate: Union[date, str] = None,
            set_country: str = None, set_address: str = None, set_phone_number: str = None, set_job_title: str = None,
    ) -> User:
        user_id = user.id if isinstance(user, User) else user
        add_identity_id = add_identity.id if isinstance(add_identity, Identity) else add_identity
        remove_identity_id = remove_identity.id if isinstance(remove_identity, Identity) else remove_identity
        payload = self.api.mutation(
            name='updateUser',
            input_value=dict(
                userId=user_id, addIdentityId=add_identity_id, removeIdentityId=remove_identity_id,
                setName=set_name, setEmail=set_email,
                setGivenName=set_given_name, setMiddleName=set_middle_name, setFamilyName=set_family_name,
                setPicture=set_picture, setBirthdate=set_birthdate,
                setCountry=set_country, setAddress=set_address, setPhoneNumber=set_phone_number,
                setJobTitle=set_job_title,
            ),
            body='user { ' + User.body() + ' }',
        )
        return User.from_dict(payload['user'])

    def delete_user(self, user: Union[User, str]) -> bool:
        user_id = user.id if isinstance(user, User) else user
        payload = self.api.mutation(
            name='deleteUser',
            input_value=dict(userId=user_id),
            body='deleted',
        )
        return bool(payload['deleted'])

    def identity(self, id: str) -> Identity:
        response = self.api.query(
            name='identity',
            variables=_filter_variables(
                ('id', 'ID!', id),
            ),
            body=Identity.body(),
        )
        return Identity.from_dict(response)

    def identities(
            self,
            provider_id: str = None,
            name_contains: str = None, email_contains: str = None,
            user: Union[User, str] = None,
            associated: bool = None, pending: bool = None, blocked: bool = None,
    ) -> list[Identity]:
        response = self.api.query(
            name='identities',
            variables=_filter_variables(
                ('providerId', 'String', provider_id),
                ('name_Icontains', 'String', name_contains),
                ('email_Icontains', 'String', email_contains),
                ('user_Id', 'String', user.id if isinstance(user, User) else user),
                ('associated', 'Boolean', associated),
                ('pending', 'Boolean', pending),
                ('blocked', 'Boolean', blocked),
            ),
            body='edges { node { ' + Identity.body() + ' } }',
        )
        return [Identity.from_dict(edge['node']) for edge in response['edges']]

    def register_identity(self, provider_id: str, id_token: str) -> Identity:
        payload = self.api.mutation(
            name='registerIdentity',
            input_value=dict(providerId=provider_id, idToken=id_token),
            body='identity { ' + Identity.body() + ' }',
        )
        return Identity.from_dict(payload['identity'])

    def update_identity(
            self, identity: Union[Identity, str],
            set_blocked: bool = None,
    ) -> Identity:
        identity_id = identity.id if isinstance(identity, Identity) else identity
        payload = self.api.mutation(
            name='updateIdentity',
            input_value=dict(identityId=identity_id, setBlocked=set_blocked),
            body='identity { ' + Identity.body() + ' }',
        )
        return Identity.from_dict(payload['identity'])

    def delete_identity(self, identity: Union[Identity, str]) -> bool:
        identity_id = identity.id if isinstance(identity, Identity) else identity
        payload = self.api.mutation(
            name='deleteIdentity',
            input_value=dict(identityId=identity_id),
            body='deleted',
        )
        return bool(payload['deleted'])

    def authorization(self, id: str) -> Authorization:
        response = self.api.query(
            name='authorization',
            variables=_filter_variables(
                ('id', 'ID!', id),
            ),
            body=Authorization.body(),
        )
        return Authorization.from_dict(response)

    def authorizations(
            self,
            name_contains: str = None,
            organization: Union[Organization, str] = None,
            member: Union[Member, str] = None,
            permission: Union[Permission, str] = None,
            role: Union[Role, str] = None,
    ) -> list[Authorization]:
        organization_id = organization.id if isinstance(organization, Organization) else organization
        member_id = member.id if isinstance(member, Member) else member
        permission_id = permission.id if isinstance(permission, Permission) else permission
        role_id = role.id if isinstance(role, Role) else role
        response = self.api.query(
            name='authorizations',
            variables=_filter_variables(
                ('organization_Id', 'String', organization_id),
                ('name_Icontains', 'String', name_contains),
                ('member_Id', 'String', member_id),
                ('permission_Id', 'String', permission_id),
                ('role_Id', 'String', role_id),
            ),
            body='edges { node { ' + Authorization.body() + ' } }',
        )
        return [Authorization.from_dict(edge['node']) for edge in response['edges']]

    def create_authorization(
            self, name: str, description: str,
            organization: Union[Organization, str],
            matching: Union[Matching, str],
            inheritance: bool,
    ) -> Authorization:
        organization_id = organization.id if isinstance(organization, Organization) else organization
        matching_value = matching.value if isinstance(matching, Matching) else matching
        payload = self.api.mutation(
            name='createAuthorization',
            input_value=dict(
                name=name, description=description,
                organizationId=organization_id,
                matching=matching_value,
                inheritance=inheritance,
            ),
            body='authorization { ' + Authorization.body() + ' }',
        )
        return Authorization.from_dict(payload['authorization'])

    def update_authorization(
            self,
            authorization: Union[Authorization, str],
            add_permission: Union[Permission, str] = None, remove_permission: Union[Permission, str] = None,
    ) -> Authorization:
        authorization_id = authorization.id if isinstance(authorization, Authorization) else authorization
        add_permission_id = add_permission.id if isinstance(add_permission, Permission) else add_permission
        remove_permission_id = remove_permission.id if isinstance(remove_permission, Permission) else remove_permission
        payload = self.api.mutation(
            name='updateAuthorization',
            input_value=dict(
                authorizationId=authorization_id,
                addPermissionId=add_permission_id,
                removePermissionId=remove_permission_id,
            ),
            body='authorization { ' + Authorization.body() + ' }',
        )
        return Authorization.from_dict(payload['authorization'])

    def delete_authorization(self, authorization: Union[Authorization, str]) -> bool:
        authorization_id = authorization.id if isinstance(authorization, Authorization) else authorization
        payload = self.api.mutation(
            name='deleteAuthorization',
            input_value=dict(authorizationId=authorization_id),
            body='deleted',
        )
        return bool(payload['deleted'])

    def role(self, id: str) -> Role:
        response = self.api.query(
            name='role',
            variables=_filter_variables(
                ('id', 'ID!', id),
            ),
            body=Role.body(),
        )
        return Role.from_dict(response)

    def roles(
            self,
            name_contains: str = None,
            organization: Union[Organization, str] = None,
            member: Union[Member, str] = None,
            authorization: Union[Authorization, str] = None,
    ) -> list[Role]:
        organization_id = organization.id if isinstance(organization, Organization) else organization
        member_id = member.id if isinstance(member, Member) else member
        authorization_id = authorization.id if isinstance(authorization, Authorization) else authorization
        response = self.api.query(
            name='roles',
            variables=_filter_variables(
                ('name_Icontains', 'String', name_contains),
                ('organization_Id', 'String', organization_id),
                ('member_Id', 'String', member_id),
                ('authorization_Id', 'String', authorization_id),
            ),
            body='edges { node { ' + Role.body() + ' } }',
        )
        return [Role.from_dict(edge['node']) for edge in response['edges']]

    def create_role(
            self, name: str, description: str,
            organization: Union[Organization, str],
            inheritance: bool,
    ) -> Role:
        organization_id = organization.id if isinstance(organization, Organization) else organization
        payload = self.api.mutation(
            name='createRole',
            input_value=dict(
                name=name, description=description,
                organizationId=organization_id,
                inheritance=inheritance,
            ),
            body='role { ' + Role.body() + ' }',
        )
        return Role.from_dict(payload['role'])

    def update_role(
            self,
            role: Union[Role, str],
            add_authorization: Union[Authorization, str] = None, remove_authorization: Union[Authorization, str] = None,
    ) -> Role:
        role_id = role.id if isinstance(role, Role) else role
        add_authorization_id = add_authorization.id if isinstance(
            add_authorization, Authorization
        ) else add_authorization
        remove_authorization_id = remove_authorization.id if isinstance(
            remove_authorization, Authorization
        ) else remove_authorization
        payload = self.api.mutation(
            name='updateRole',
            input_value=dict(
                roleId=role_id,
                addAuthorizationId=add_authorization_id,
                removeAuthorizationId=remove_authorization_id,
            ),
            body='role { ' + Role.body() + ' }',
        )
        return Role.from_dict(payload['role'])

    def delete_role(self, role: Union[Role, str]) -> bool:
        role_id = role.id if isinstance(role, Role) else role
        payload = self.api.mutation(
            name='deleteRole',
            input_value=dict(roleId=role_id),
            body='deleted',
        )
        return bool(payload['deleted'])

    def member(self, id: str) -> Member:
        response = self.api.query(
            name='member',
            variables=_filter_variables(
                ('id', 'ID!', id),
            ),
            body=Member.body(),
        )
        return Member.from_dict(response)

    def members(self, user: Union[User, str] = None, organization: Union[Organization, str] = None) -> list[Member]:
        user_id = user.id if isinstance(user, User) else user
        organization_id = organization.id if isinstance(organization, Organization) else organization
        response = self.api.query(
            name='members',
            variables=_filter_variables(
                ('user_Id', 'String', user_id),
                ('organization_Id', 'String', organization_id),
            ),
            body='edges { node { ' + Member.body() + ' } }',
        )
        return [Member.from_dict(edge['node']) for edge in response['edges']]

    def create_member(
            self,
            user: Union[User, str],
            organization: Union[Organization, str],
            uuid: str = None,
    ) -> Member:
        user_id = user.id if isinstance(user, User) else user
        organization_id = organization.id if isinstance(organization, Organization) else organization
        payload = self.api.mutation(
            name='createMember',
            input_value=dict(
                uuid=uuid,
                userId=user_id,
                organizationId=organization_id,
            ),
            body='member { ' + Member.body() + ' }',
        )
        return Member.from_dict(payload['member'])

    def update_member(
            self,
            member: Union[Member, str],
            add_authorization: Union[Authorization, str] = None, remove_authorization: Union[Authorization, str] = None,
            add_role: Union[Role, str] = None, remove_role: Union[Role, str] = None,
    ) -> Member:
        member_id = member.id if isinstance(member, Member) else member
        add_authorization_id = add_authorization.id if isinstance(
            add_authorization, Authorization
        ) else add_authorization
        remove_authorization_id = remove_authorization.id if isinstance(
            remove_authorization, Authorization
        ) else remove_authorization
        add_role_id = add_role.id if isinstance(add_role, Role) else add_role
        remove_role_id = remove_role.id if isinstance(remove_role, Role) else remove_role
        payload = self.api.mutation(
            name='updateMember',
            input_value=dict(
                memberId=member_id,
                addAuthorizationId=add_authorization_id,
                removeAuthorizationId=remove_authorization_id,
                addRoleId=add_role_id,
                removeRoleId=remove_role_id,
            ),
            body='member { ' + Member.body() + ' }',
        )
        return Member.from_dict(payload['member'])

    def delete_member(self, member: Union[Member, str]) -> bool:
        member_id = member.id if isinstance(member, Member) else member
        payload = self.api.mutation(
            name='deleteMember',
            input_value=dict(memberId=member_id),
            body='deleted',
        )
        return bool(payload['deleted'])

    def account(self, id: str) -> Account:
        response = self.api.query(
            name='account',
            variables=_filter_variables(
                ('id', 'ID!', id),
            ),
            body=Account.body(),
        )
        return Account.from_dict(response)

    def accounts(
            self,
            application: Union[Application, str] = None,
            user: Union[User, str] = None,
            enabled: bool = None,
    ) -> list[Account]:
        application_id = application.id if isinstance(application, Application) else application
        user_id = user.id if isinstance(user, User) else user
        response = self.api.query(
            name='accounts',
            variables=_filter_variables(
                ('application_Id', 'String', application_id),
                ('user_Id', 'String', user_id),
                ('enabled', 'Boolean', enabled),
            ),
            body='edges { node { ' + Account.body() + ' } }',
        )
        return [Account.from_dict(edge['node']) for edge in response['edges']]

    def create_account(self, user: Union[User, str], application: Union[Application, str]) -> Account:
        user_id = user.id if isinstance(user, User) else user
        application_id = application.id if isinstance(application, Application) else application
        payload = self.api.mutation(
            name='createAccount',
            input_value=dict(userId=user_id, applicationId=application_id),
            body='account { ' + Account.body() + ' }',
        )
        return Account.from_dict(payload['account'])

    def update_account(
            self, account: Union[Account, str],
            set_enabled: bool = None,
    ) -> Account:
        account_id = account.id if isinstance(account, Account) else account
        payload = self.api.mutation(
            name='updateAccount',
            input_value=dict(
                accountId=account_id,
                setEnabled=set_enabled,
            ),
            body='account { ' + Account.body() + ' }',
        )
        return Account.from_dict(payload['account'])

    def delete_account(self, account: Union[Account, str]) -> bool:
        account_id = account.id if isinstance(account, Account) else account
        payload = self.api.mutation(
            name='deleteAccount',
            input_value=dict(accountId=account_id),
            body='deleted',
        )
        return bool(payload['deleted'])

    def create_user_request(
            self,
            identity: Union[Identity, str],
            name: str = None,
            given_name: str = None,
            middle_name: str = None,
            family_name: str = None,
            picture: str = None,
            birthdate: Union[date, str] = None,
            country: str = None,
            address: str = None,
            phone_number: str = None,
            job_title: str = None,
            comment: str = None,
    ) -> UserRequest:
        identity_id = identity.id if isinstance(identity, Identity) else identity
        birthdate = date.fromisoformat(birthdate) if isinstance(birthdate, str) else birthdate
        payload = self.api.mutation(
            name='createUserRequest',
            input_value=dict(
                identityId=identity_id, comment=comment,
                name=name,
                givenName=given_name,
                middleName=middle_name,
                familyName=family_name,
                picture=picture,
                birthdate=birthdate.isoformat() if birthdate else None,
                country=country,
                address=address,
                phoneNumber=phone_number,
                jobTitle=job_title,
            ),
            body='userRequest { ' + UserRequest.body() + ' }',
        )
        return UserRequest.from_dict(payload['userRequest'])

    def accept_user_request(self, user_request: Union[UserRequest, str]) -> tuple[User, Identity]:
        user_request_id = user_request.id if isinstance(user_request, UserRequest) else user_request
        payload = self.api.mutation(
            name='acceptUserRequest',
            input_value=dict(userRequestId=user_request_id),
            body='user { ' + User.body() + ' } identity { ' + Identity.body() + ' }',
        )
        return User.from_dict(payload['user']), Identity.from_dict(payload['identity'])

    def reject_user_request(
            self, user_request: Union[UserRequest, str],
            block_identity: bool = False,
            delete_identity: bool = False,
    ) -> Optional[Identity]:
        user_request_id = user_request.id if isinstance(user_request, UserRequest) else user_request
        payload = self.api.mutation(
            name='rejectUserRequest',
            input_value=dict(
                userRequestId=user_request_id,
                blockIdentity=block_identity,
                deleteIdentity=delete_identity,
            ),
            body='identity { ' + Identity.body() + ' }',
        )
        identity = Identity.from_dict(payload['identity']) if payload['identity'] else None
        return identity

    def user_request(self, id: str) -> UserRequest:
        response = self.api.query(
            name='userRequest',
            variables=_filter_variables(
                ('id', 'ID!', id),
            ),
            body=UserRequest.body(),
        )
        return UserRequest.from_dict(response)

    def user_requests(
            self,
            identity: Union[Identity, str] = None,
    ) -> list[UserRequest]:
        identity_id = identity.id if isinstance(identity, Identity) else identity
        response = self.api.query(
            name='userRequests',
            variables=_filter_variables(
                ('identity_Id', 'String', identity_id),
            ),
            body='edges { node { ' + UserRequest.body() + ' } }',
        )
        return [UserRequest.from_dict(edge['node']) for edge in response['edges']]

    def ticket_for_identity(self, identity: Union[Identity, str]) -> str:
        identity_id = identity.id if isinstance(identity, Identity) else identity
        response: str = self.api.query(
            name='identityTicket',
            variables=_filter_variables(
                ('identityId', 'ID!', identity_id),
            ),
        )
        return response

    def identity_for_ticket(self, ticket: str) -> Optional[Identity]:
        try:
            identity_id: str = self.api.query(
                name='checkIdentityTicket',
                variables=_filter_variables(
                    ('identityTicket', 'String!', ticket),
                ),
            )
        except Exception:
            return None
        return self.identity(identity_id)

    def decode_token(self, token: str) -> TokenInfo:
        auth_metadata = self.api.auth_metadata()
        if not token:
            return TokenInfo()
        token_algorithm = auth_metadata['token_algorithm']
        token_public_key = auth_metadata['token_public_key']
        try:
            decoded = jwt.decode(token, token_public_key, algorithms=[token_algorithm])
        except Exception:
            return TokenInfo()
        return TokenInfo(
            user_id=decoded.get('sub'),
            display_name=decoded.get('metadata', {}).get('displayed_name'),
            given_name=decoded.get('metadata', {}).get('given_name'),
            middle_name=decoded.get('metadata', {}).get('middle_name'),
            family_name=decoded.get('metadata', {}).get('family_name'),
            email=decoded.get('metadata', {}).get('email'),
            is_superuser=decoded.get('is_superuser'),
            member_id=decoded.get('metadata', {}).get('member_id'),
            organization_id=decoded.get('metadata', {}).get('organization_id'),
            organization_name=decoded.get('metadata', {}).get('organization_name'),
            organization_namespace=decoded.get('metadata', {}).get('organization_namespace'),
            applications=decoded.get('applications'),
        )

    def actor(self, token: str = None, secret: str = None) -> Actor:
        if not token or not secret:
            return Actor()
        response = self.api.query(
            name='actor',
            variables=_filter_variables(
                ('tokenCookie', 'String', token),
                ('secretCookie', 'String', secret),
            ),
            body=Actor.body(),
        )
        return Actor.from_dict(response)

    def cookies(self, user: Union[User, str] = None, organization: Union[Organization, str] = None) -> CookiePair:
        user_id = user.id if isinstance(user, User) else user
        organization_id = organization.id if isinstance(organization, Organization) else organization
        response = self.api.query(
            name='cookies',
            variables=_filter_variables(
                ('userId', 'ID', user_id),
                ('organizationId', 'ID', organization_id),
            ),
            body=CookiePair.body(),
        )
        return CookiePair.from_dict(response)

    def is_authorized(
            self,
            actor: Union[User, Member, Actor, str],
            permission: Union[Permission, str],
            organization: Union[Organization, Namespace, str] = None,
    ) -> bool:
        actor_id = actor
        if isinstance(actor, User):
            actor_id = actor.id
        elif isinstance(actor, Member):
            actor_id = actor.id
        elif isinstance(actor, Actor):
            if actor.member:
                actor_id = actor.member.id
            elif actor.user:
                actor_id = actor.user.id
            else:
                return False
        permission_id = permission.id if isinstance(permission, Permission) else permission
        organization_id = organization
        if isinstance(organization, Organization):
            organization_id = organization.id
        elif isinstance(organization, Namespace):
            organization_id = str(organization)

        response = self.api.query(
            name='isAuthorized',
            variables=_filter_variables(
                ('actorId', 'ID!', actor_id),
                ('permissionId', 'ID!', permission_id),
                ('organizationId', 'String', organization_id),
            ),
        )
        return bool(response)

    def organizations_where_is_authorized(
            self,
            actor: Union[User, Member, Actor, str],
            permission: Union[Permission, str],
    ) -> list[Organization]:
        actor_id = actor
        if isinstance(actor, User):
            actor_id = actor.id
        elif isinstance(actor, Member):
            actor_id = actor.id
        elif isinstance(actor, Actor):
            if actor.member:
                actor_id = actor.member.id
            elif actor.user:
                actor_id = actor.user.id
            else:
                return []
        permission_id = permission.id if isinstance(permission, Permission) else permission

        response = self.api.query(
            name='organizationsWhereIsAuthorized',
            variables=_filter_variables(
                ('actorId', 'ID!', actor_id),
                ('permissionId', 'ID!', permission_id),
            ),
            body=Organization.body(),
        )
        return [Organization.from_dict(org) for org in response]

    def last_change_date(self) -> datetime:
        response = self.api.query(
            name='lastChangeDate',
        )
        return datetime.fromisoformat(response)



# Copyright (c) Qotto, 2022

from __future__ import annotations

import dataclasses
import re
import typing
from dataclasses import dataclass
from enum import Enum
from typing import Union, Optional

__all__ = [
    'Namespace',
    'Matching',
    'Application',
    'Permission',
    'Organization',
    'Authorization',
    'Role',
    'User',
    'Member',
    'Actor',
    'Account',
    'Cookie',
    'Identity',
    'UserRequest',
    'CookiePair',
    'TokenInfo',
]


class Matching(Enum):
    ALL = 'ALL'
    EXACT = 'EXACT'
    ASCENDANT = 'ASCENDANT'
    DESCENDANT = 'DESCENDANT'
    SAME_BRANCH = 'SAME_BRANCH'

    def __str__(self) -> str:
        return self.name


class Namespace:
    _nodes: list[str]

    def __init__(
            self,
            nodes: Union[str, list[str]],
    ) -> None:
        """
        >>> Namespace('a:::b  :c:d')
        a:b:c:d
        >>> Namespace(['a', '  b  ', 'c', '', '', 'd'])
        a:b:c:d
        >>> Namespace(['a:', 'b', 'c', 'd'])
        Traceback (most recent call last):
            ...
        ValueError: A node cannot contain colons ":".
        >>> Namespace(['a', 1, 'c', 'd'])
        Traceback (most recent call last):
            ...
        TypeError: A node must be a str.
        >>> Namespace('')
        Traceback (most recent call last):
            ...
        ValueError: Namespace must have at least one node.
        """
        if isinstance(nodes, str):
            nodes = list(filter(len, (v.strip().lower() for v in nodes.split(':'))))
        elif isinstance(nodes, list):
            if any(filter(lambda x: not isinstance(x, str), nodes)):
                raise TypeError("A node must be a str.")
            if any(filter(lambda x: ':' in str(x), nodes)):
                raise ValueError("A node cannot contain colons \":\".")
            nodes = list(filter(len, (v.strip().lower() for v in nodes)))
        if not nodes:
            raise ValueError("Namespace must have at least one node.")
        for node in nodes:
            if not node:
                raise ValueError("Namespace node must not be empty.")
        self._nodes = nodes

    @property
    def path(self) -> str:
        return ':'.join(self._nodes)

    @property
    def nodes(self) -> list[str]:
        return self._nodes.copy()

    def matches(self, other: Namespace, matching: Matching = Matching.ALL) -> bool:
        """
        Test is self is <matching> of other.
        """
        if matching == Matching.ALL:
            return True
        if matching == Matching.EXACT:
            return self == other
        if matching == Matching.ASCENDANT:
            return self >= other
        if matching == Matching.DESCENDANT:
            return self <= other
        if matching == Matching.SAME_BRANCH:
            return self <= other or self >= other
        raise ValueError(f"Unknown matching type {matching}.")

    def __eq__(self, other) -> bool:
        return (isinstance(other, Namespace)
                and len(self) == len(other)
                and self.path == other.path)

    def __hash__(self):
        return hash(self.path)

    def __le__(self, other) -> bool:
        return (isinstance(other, Namespace)
                and len(self) >= len(other)
                and Namespace(self._nodes[:len(other)]) == other)

    def __ge__(self, other) -> bool:
        return (isinstance(other, Namespace)
                and len(self) <= len(other)
                and Namespace(other._nodes[:len(self)]) == self)

    def __lt__(self, other) -> bool:
        return (isinstance(other, Namespace)
                and len(self) > len(other)
                and Namespace(self._nodes[:len(other)]) == other)

    def __gt__(self, other) -> bool:
        return (isinstance(other, Namespace)
                and len(self) < len(other)
                and Namespace(other._nodes[:len(self)]) == self)

    def __len__(self) -> int:
        return len(self._nodes)

    def __str__(self) -> str:
        return self.path

    def __repr__(self) -> str:
        return self.path


_camel_pattern = re.compile(r"([a-zA-Z][^A-Z0-9]*|[0-9]+)")


def _camel_case(snake_case: str) -> str:
    components = snake_case.split('_')
    return components[0] + ''.join(x.title() for x in components[1:])


def _snake_case(camel_case: str) -> str:
    components = _camel_pattern.findall(camel_case)
    return "_".join(component.lower() for component in components)


Self = typing.TypeVar("Self", bound="GqlData")


@dataclass(frozen=True)
class GqlData:
    @classmethod
    def from_dict(cls, data: dict) -> Self:
        kwargs = dict()
        type_hints = typing.get_type_hints(cls)
        for field in dataclasses.fields(cls):
            field_name = field.name
            field_type = type_hints[field_name]
            value = None
            if field_name in data:
                value = data[field_name]
            elif _camel_case(field_name) in data:
                value = data[_camel_case(field_name)]
            else:
                continue
            if not isinstance(field_type, type):
                for optional_type in field_type.__args__:
                    if issubclass(optional_type, type(None)):
                        continue
                    field_type = optional_type
            if issubclass(field_type, GqlData) and value is not None:
                value = field_type.from_dict(value)
            elif issubclass(field_type, Matching):
                value = Matching(value)
            elif issubclass(field_type, Namespace):
                value = Namespace(value)
            kwargs[field_name] = value
        return cls(**kwargs)  # type: ignore

    def to_dict(self, capitalize: bool = True) -> dict:
        result = dict()
        for field in dataclasses.fields(self):
            field_name = field.name
            key = _camel_case(field_name) if capitalize else field_name
            value = getattr(self, field_name)
            if isinstance(value, GqlData):
                value = value.to_dict(capitalize)
            elif isinstance(value, Namespace):
                value = str(value)
            elif isinstance(value, Matching):
                value = value.value
            result[key] = value
        return result

    @classmethod
    def body(cls) -> str:
        result = list()
        type_hints = typing.get_type_hints(cls)
        for field in dataclasses.fields(cls):
            field_name = _camel_case(field.name)
            field_type = type_hints[field.name]
            result.append(field_name)
            if not isinstance(field_type, type):
                for optional_type in field_type.__args__:
                    if issubclass(optional_type, type(None)):
                        continue
                    field_type = optional_type
            if issubclass(field_type, GqlData):
                result.append('{ ' + field_type.body() + ' }')
        return ' '.join(result)

    def update(self, **kwargs) -> Self:
        data = self.to_dict()
        for k, v in kwargs.items():
            if isinstance(v, GqlData):
                v = v.to_dict()
            data[_camel_case(k)] = v
        return self.from_dict(data)


@dataclass(frozen=True)
class Application(GqlData):
    id: str
    code: str
    name: str
    description: str


@dataclass(frozen=True)
class Permission(GqlData):
    id: str
    application: Application
    code: str
    name: str
    description: str


@dataclass(frozen=True)
class Organization(GqlData):
    id: str
    name: str
    namespace: Namespace


@dataclass(frozen=True)
class Authorization(GqlData):
    id: str
    name: str
    description: str
    organization: Organization
    inheritance: bool
    matching: Matching


@dataclass(frozen=True)
class Role(GqlData):
    id: str
    name: str
    description: str
    organization: Organization
    inheritance: bool


@dataclass(frozen=True)
class Identity(GqlData):
    id: str
    name: str
    provider_id: str
    email: str
    user: Optional[User]
    user_request: Optional[UserRequest]
    blocked: bool


@dataclass(frozen=True)
class User(GqlData):
    id: str
    name: str
    is_superuser: bool


@dataclass(frozen=True)
class Member(GqlData):
    id: str
    user: User
    organization: Organization


@dataclass(frozen=True)
class Account(GqlData):
    id: str
    application: Application
    user: User
    enabled: bool
    data: dict


@dataclass(frozen=True)
class UserRequest(GqlData):
    id: str
    name: str
    comment: Optional[str] = None


@dataclass(frozen=True)
class Actor(GqlData):
    user: Optional[User] = None
    member: Optional[Member] = None


@dataclass(frozen=True)
class Cookie(GqlData):
    name: str
    value: str
    domain: str
    max_age: int
    secure: bool
    http_only: bool

    def __str__(self):
        return f'Cookie"{self.name}={self.value}"'


@dataclass(frozen=True)
class CookiePair(GqlData):
    token_cookie: Cookie
    secret_cookie: Cookie


@dataclass(frozen=True)
class TokenInfo:
    user_id: Optional[str] = None
    display_name: Optional[str] = None
    given_name: Optional[str] = None
    middle_name: Optional[str] = None
    family_name: Optional[str] = None
    email: Optional[str] = None
    is_superuser: Optional[bool] = None
    member_id: Optional[str] = None
    organization_id: Optional[str] = None
    organization_name: Optional[str] = None
    organization_namespace: Optional[str] = None
    applications: Optional[list[str]] = None



from qottoauth.api.base import QottoAuthApi, QottoAuthApiError
from qottoauth.api.gql import QottoAuthGqlApi
from qottoauth.api.test import QottoAuthTestApi
from qottoauth.api.cached import QottoAuthCachedApi

__all__ = [
    'QottoAuthApi',
    'QottoAuthApiError',
    'QottoAuthTestApi',
    'QottoAuthGqlApi',
    'QottoAuthCachedApi',
]



from datetime import datetime, timezone
from typing import Any, Tuple, Optional
from uuid import uuid4

import jwt

from qottoauth import Namespace, Matching
from qottoauth.api.base import QottoAuthApi, QottoAuthApiError
from qottoauth.models import (
    Role,
    Organization,
    Identity,
    Account,
    Application,
    Permission,
    Authorization,
    Member,
    User,
    UserRequest,
    Actor,
    Cookie,
    CookiePair,
)

__all__ = [
    'QottoAuthTestApi',
]


def gen_uuid():
    return str(uuid4())


class QottoAuthTestApi(QottoAuthApi):

    def __init__(self):
        self._applications: dict[str, Application] = {}
        self._accounts: dict[str, Account] = {}
        self._permissions: dict[str, Permission] = {}
        self._authorizations: dict[str, Authorization] = {}
        self._authorization_permissions: dict[str, set[str]] = {}
        self._roles: dict[str, Role] = {}
        self._role_authorizations: dict[str, set[str]] = {}
        self._members: dict[str, Member] = {}
        self._member_roles: dict[str, set[str]] = {}
        self._member_authorizations: dict[str, set[str]] = {}
        self._organizations: dict[str, Organization] = {}
        self._users: dict[str, User] = {}
        self._identities: dict[str, Identity] = {}
        self._identity_tickets: dict[str, str] = {}
        self._user_requests: dict[str, UserRequest] = {}
        self._last_change_date: datetime = datetime.now(timezone.utc)

        self._token_cookie_name = 'token'
        self._secret_cookie_name = 'secret'
        self._token_algorithm = 'RS256'
        self._secret_algorithm = 'RS256'
        self._token_public_key = '''-----BEGIN RSA PUBLIC KEY-----
MIGJAoGBAMvJcGqEd7hZL+LFBehqj73rqJOpT5A/yWBfcrpTTgq7hV0rthGHLiQM
i5U7zjT+nbxYmCytig499egpibTPCnU/0SqXmlx4jelQb4QK30/TdktESlhgb5dg
2qC5IhBrii/pb36wO+Tjz3t1F9EqWxnMRje3qdhVOJr/J9cdl/GDAgMBAAE=
-----END RSA PUBLIC KEY-----'''
        self._token_private_key = '''-----BEGIN RSA PRIVATE KEY-----
MIICXAIBAAKBgQDLyXBqhHe4WS/ixQXoao+966iTqU+QP8lgX3K6U04Ku4VdK7YR
hy4kDIuVO840/p28WJgsrYoOPfXoKYm0zwp1P9Eql5pceI3pUG+ECt9P03ZLREpY
YG+XYNqguSIQa4ov6W9+sDvk4897dRfRKlsZzEY3t6nYVTia/yfXHZfxgwIDAQAB
AoGALSTEqHQL6WSofP3UXzYr/ZCU2ZEqNPRMcfLzAV+u7CW6V3d1b9CYYhf9W+PR
vF+jJbQikdMnwBvtu65n+QvcbtMKpXcmcRuaHqs4x7IoSxNKb1ILz+do1ekrFc16
5oPd0EvGLyFcjH7O61He0iY3Uf/tRcKiCUPGU4BLbvLocZECQQD/ttxNFGCxxv2/
ojiS4Ii1ODJ8QtbdKBp2fKUOf8XgvgHgt5KuvVW8kf6lBMqWcwNt85QskTDNLtCI
r9dxJcuHAkEAzAO57aIEEDc0Qvp0gKK+34MRQtfBgAWNLF5amBtKdUd4B9QMdYbr
fhOovbKtVca4ra93SRS01InkTV3hKXoBJQJAbaE3B7DB19XpOfxRZt7unUrvkgiR
15T262960CGFc1nisjXhpBq2JDcvRg4s0J2UjdIM56KDmqQEcWV55x9+BwJAYVrd
3OJVog9V5yhxc/k1sJ9xGz6uXhNIHQYhoThUvcaPJt0v3N23fwCOo4eiY65i7q8u
8zboXAw5YBoOmqZX6QJBANyXY4RBSZC1xxJiWB6YGezRf4H4jR7+4ZfYDNfjBdcv
IuFD3aYARUKtdS1jZW1Si5Bbhw7lLxxm9qgcmlZq00A=
-----END RSA PRIVATE KEY-----'''
        self._secret_private_key = self._token_private_key

    def query(
            self,
            name: str,
            variables: list[Tuple[str, str, Any]] = None,
            body: str = None,
    ):
        kwargs = {}
        if variables:
            for var_name, var_type, var_value in variables:
                kwargs[var_name] = var_value
        if hasattr(self, name):
            try:
                return getattr(self, name)(**kwargs)
            except Exception as e:
                raise QottoAuthApiError from e
        raise QottoAuthApiError(f"Unknown query {name}")

    def mutation(
            self,
            name: str,
            body: str,
            input_name: str = 'input',
            input_type: str = None,
            input_value: dict[str, Any] = None,
    ):
        if hasattr(self, name):
            try:
                result = getattr(self, name)(**input_value)
                self._last_change_date = datetime.now(timezone.utc)
                return result
            except Exception as e:
                raise QottoAuthApiError from e
        raise QottoAuthApiError(f"Unknown mutation {name}")

    def auth_metadata(self) -> dict[str, str]:
        return dict(
            token_cookie_name=self._token_cookie_name,
            token_public_key=self._token_public_key,
            token_algorithm=self._token_algorithm,
            secret_cookie_name=self._secret_cookie_name,
        )

    def lastChangeDate(self) -> str:
        return self._last_change_date.isoformat()

    def application(self, id) -> dict:
        if id in self._applications:
            return self._applications[id].to_dict()
        raise QottoAuthApiError(f"Application {id} does not exists")

    def applications(self, name_Icontains: str = None) -> dict:
        nodes = []
        for application in self._applications.values():
            if name_Icontains is None or name_Icontains.lower() in application.name.lower():
                nodes.append(application.to_dict())
        return dict(edges=[dict(node=node) for node in nodes])

    def createApplication(self, code: str, name: str = None, description: str = None, uuid: str = None,
                          **kwargs) -> dict:
        uuid = uuid or gen_uuid()
        name = name or code
        code = code or name
        description = description or name
        if not code:
            raise QottoAuthApiError("Code is required")
        for application in self._applications.values():
            if application.code == code:
                raise QottoAuthApiError(f"Application {code} already exists")
        application = Application(uuid, code, name, description)
        self._applications[application.id] = application
        return dict(application=application.to_dict())

    def registerApplication(
            self,
            code: str, name: str = None, description: str = None, uuid: str = None, **kwargs,
    ) -> dict:
        uuid = uuid or gen_uuid()
        name = name or code
        code = code or name
        description = description or name
        if not code:
            raise QottoAuthApiError("Code is required")
        for application in self._applications.values():
            if application.code == code:
                updated_application: Application = application.update(name=name, description=description)
                self._applications[application.id] = updated_application
                return dict(application=updated_application.to_dict())
        new_application = Application(uuid, code, name, description)
        self._applications[new_application.id] = new_application
        return dict(application=new_application.to_dict())

    def deleteApplication(self, applicationId: str) -> dict:
        if applicationId in self._applications:
            del self._applications[applicationId]
            for permission in list(self._permissions.values()):
                if permission.application.id == applicationId:
                    self.deletePermission(permission.id)
            for account in list(self._accounts.values()):
                if account.application.id == applicationId:
                    self.deleteAccount(account.id)
            return dict(deleted=True)
        return dict(deleted=False)

    def permission(self, id) -> dict:
        if id in self._permissions:
            return self._permissions[id].to_dict()
        raise QottoAuthApiError(f"Permission {id} does not exists")

    def permissions(
            self, name_Icontains: str = None,
            application_Id: str = None, authorization_Id: str = None
    ) -> dict:
        nodes = []
        for permission in self._permissions.values():
            if name_Icontains is not None and name_Icontains.lower() not in permission.name.lower():
                continue
            if application_Id is not None and application_Id != permission.application.id:
                continue
            if authorization_Id is not None and permission.id not in self._authorization_permissions.get(
                    authorization_Id, []
            ):
                continue
            nodes.append(permission.to_dict())
        return dict(edges=[dict(node=node) for node in nodes])

    def createPermission(
            self,
            applicationId: str, code: str, name: str = None, description: str = None, uuid: str = None, **kwargs,
    ) -> dict:
        application = self._applications[applicationId]
        uuid = uuid or gen_uuid()
        name = name or code
        code = code or name
        description = description or name
        if not code:
            raise QottoAuthApiError("Code is required")
        for permission in self._permissions.values():
            if permission.code == name and permission.application.id == application.id:
                raise QottoAuthApiError(f"Permission {code} already exists for this application")
        permission = Permission(uuid, application, code, name, description)
        self._permissions[permission.id] = permission
        return dict(permission=permission.to_dict())

    def registerPermission(
            self,
            applicationId: str, code: str, name: str = None, description: str = None, uuid=None, **kwargs,
    ) -> dict:
        application = self._applications[applicationId]
        uuid = uuid or gen_uuid()
        name = name or code
        code = code or name
        description = description or name
        if not code:
            raise QottoAuthApiError("Code is required")
        for permission in self._permissions.values():
            if permission.code == code and permission.application.id == application.id:
                updated_permission: Permission = permission.update(name=name, description=description)
                self._permissions[permission.id] = updated_permission
                return dict(permission=updated_permission.to_dict())
        new_permission = Permission(uuid, application, code, name, description)
        self._permissions[new_permission.id] = new_permission
        return dict(permission=new_permission.to_dict())

    def registerPermissions(
            self,
            applicationId: str, permissions: list[dict[str, Any]], **kwargs,
    ) -> dict:
        result = []
        for permission_info in permissions:
            result.append(self.registerPermission(applicationId, **permission_info)['permission'])
        return dict(permissions=result)

    def deletePermission(self, permissionId: str) -> dict:
        if permissionId in self._permissions:
            del self._permissions[permissionId]
            for authorization_id, permission_id_set in self._authorization_permissions.items():
                permission_id_set.discard(permissionId)
            return dict(deleted=True)
        return dict(deleted=False)

    def organization(self, id: str) -> dict:
        if id in self._organizations:
            return self._organizations[id].to_dict()
        raise QottoAuthApiError(f"Organization {id} does not exists")

    def organizations(self, name_Icontains: str = None, namespace_Icontains: str = None) -> dict:
        nodes = []
        for organization in self._organizations.values():
            if name_Icontains is None or name_Icontains.lower() in organization.name.lower():
                if namespace_Icontains is None or namespace_Icontains.lower() in str(organization.namespace).lower():
                    nodes.append(organization.to_dict())
        return dict(edges=[dict(node=node) for node in nodes])

    def createOrganization(self, namespace: str, name: str = None, uuid: str = None, **kwargs) -> dict:
        uuid = uuid or gen_uuid()
        name = name or namespace
        for organization in self._organizations.values():
            if organization.namespace == namespace:
                raise QottoAuthApiError(f"Organization {namespace} already exists")
        organization = Organization(uuid, name, Namespace(namespace))
        self._organizations[organization.id] = organization
        return dict(organization=organization.to_dict())

    def registerOrganization(self, namespace: str, name: str = None, uuid: str = None, **kwargs) -> dict:
        uuid = uuid or gen_uuid()
        name = name or namespace
        for organization in self._organizations.values():
            if str(organization.namespace) == namespace:
                updated_organization: Organization = organization.update(name=name)
                self._organizations[organization.id] = updated_organization
                return dict(organization=updated_organization.to_dict())
        organization = Organization(uuid, name, Namespace(namespace))
        self._organizations[organization.id] = organization
        return dict(organization=organization.to_dict())

    def deleteOrganization(self, organizationId: str) -> dict:
        if organizationId in self._organizations:
            del self._organizations[organizationId]
            for member in list(self._members.values()):
                if member.organization.id == organizationId:
                    self.deleteMember(member.id)
            for authorization in list(self._authorizations.values()):
                if authorization.organization.id == organizationId:
                    self.deleteAuthorization(authorization.id)
            for role in list(self._roles.values()):
                if role.organization.id == organizationId:
                    self.deleteRole(role.id)
            return dict(deleted=True)
        return dict(deleted=False)

    def user(self, id: str) -> dict:
        if id in self._users:
            return self._users[id].to_dict()
        raise QottoAuthApiError(f"User {id} does not exists")

    def users(self, name_Icontains: str = None, associated: bool = None) -> dict:
        nodes = []
        for user in self._users.values():
            if name_Icontains is not None and name_Icontains.lower() not in user.name.lower():
                continue
            if associated is not None:
                for identity in self._identities.values():
                    if identity.user and identity.user.id == user.id:
                        has_identity = True
                        break
                else:
                    has_identity = False
                if associated != has_identity:
                    continue
            nodes.append(user.to_dict())
        return dict(edges=[dict(node=node) for node in nodes])

    def createUser(self, name: str, uuid: str = None, **kwargs) -> dict:
        uuid = uuid or gen_uuid()
        is_superuser = kwargs.get('isSuperuser') or False
        user = User(uuid, name, is_superuser)
        self._users[user.id] = user
        return dict(user=user.to_dict())

    def updateUser(
            self, userId: str, addIdentityId: str = None, removeIdentityId: str = None,
            **kwargs,
    ) -> dict:
        user = self._users[userId]
        if addIdentityId:
            add_identity = self._identities[addIdentityId]
            if add_identity.user:
                raise QottoAuthApiError(f"Identity {addIdentityId} already belongs to user {add_identity.user.id}")
            self._identities[addIdentityId] = add_identity.update(user=user)
        if removeIdentityId:
            remove_identity = self._identities[removeIdentityId]
            if not remove_identity.user or remove_identity.user.id != userId:
                raise QottoAuthApiError(f"Identity {removeIdentityId} does not belongs to user {userId}")
            self._identities[removeIdentityId] = remove_identity.update(user=None)
        return dict(user=user.to_dict())

    def createUserFromIdentity(self, identityId: str) -> dict:
        identity = self._identities[identityId]
        user = User(gen_uuid(), identity.name, False)
        self._users[user.id] = user
        self._identities[identityId] = identity.update(user=user)
        return dict(user=user.to_dict())

    def deleteUser(self, userId: str) -> dict:
        if userId in self._users:
            del self._users[userId]
            for member in list(self._members.values()):
                if member.user.id == userId:
                    self.deleteMember(member.id)
            for account in list(self._accounts.values()):
                if account.user.id == userId:
                    self.deleteAccount(account.id)
            return dict(deleted=True)
        return dict(deleted=False)

    def identity(self, id: str) -> dict:
        if id in self._identities:
            identity = self._identities[id]
            return identity.to_dict()
        raise QottoAuthApiError(f"Identity {id} does not exists")

    def identities(
            self,
            name_Icontains: str = None, email_Icontains: str = None,
            providerId: str = None,
            associated: bool = None, user_Id: str = None,
            blocked: bool = None,
    ) -> dict:
        nodes = []
        for identity in self._identities.values():
            if providerId is not None and identity.provider_id != providerId:
                continue
            if name_Icontains is not None and name_Icontains.lower() not in identity.name.lower():
                continue
            if email_Icontains is not None and email_Icontains.lower() not in identity.email.lower():
                continue
            if associated is not None and associated != (identity.user is not None):
                continue
            if user_Id is not None and (identity.user is None or identity.user.id != user_Id):
                continue
            if blocked is not None and blocked != identity.blocked:
                continue
            nodes.append(identity.to_dict())
        return dict(edges=[dict(node=node) for node in nodes])

    def registerIdentity(self, providerId: str, idToken: str) -> dict:
        if providerId != 'dummy':
            raise QottoAuthApiError(f"Provider {providerId} does not exists.")
        identity_uuid, identity_name, identity_email = idToken.split(':')
        if identity_uuid not in self._identities:
            self._identities[identity_uuid] = Identity(
                id=identity_uuid, name=identity_name, provider_id=providerId,
                email=identity_email, user=None, blocked=False, user_request=None,
            )
            self._identity_tickets[identity_uuid] = str(uuid4())
        return dict(identity=self._identities[identity_uuid].to_dict())

    def updateIdentity(self, identityId: str, setBlocked: bool = None) -> dict:
        identity = self._identities[identityId]
        if setBlocked is not None and setBlocked != identity.blocked:
            identity = identity.update(blocked=setBlocked)
            self._identities[identityId] = identity
        return dict(identity=identity.to_dict())

    def deleteIdentity(self, identityId: str) -> dict:
        if identityId in self._identities:
            user_request = self._identities[identityId].user_request
            if user_request and user_request.id in self._user_requests:
                del self._user_requests[user_request.id]
            del self._identities[identityId]
            del self._identity_tickets[identityId]
            return dict(deleted=True)
        return dict(deleted=False)

    def authorization(self, id: str) -> dict:
        if id in self._authorizations:
            return self._authorizations[id].to_dict()
        raise QottoAuthApiError(f"Authorization {id} does not exists")

    def authorizations(
            self, name_Icontains: str = None,
            organization_Id: str = None,
            member_Id: str = None, permission_Id: str = None, role_Id: str = None,
    ) -> dict:
        nodes = []
        for authorization in self._authorizations.values():
            if name_Icontains is not None and name_Icontains.lower() not in authorization.name.lower():
                continue
            if organization_Id is not None and authorization.organization.id != organization_Id:
                continue
            if member_Id is not None and authorization.id not in self._member_authorizations[member_Id]:
                continue
            if permission_Id is not None and permission_Id not in self._authorization_permissions[authorization.id]:
                continue
            if role_Id is not None and authorization.id not in self._role_authorizations[role_Id]:
                continue
            nodes.append(authorization.to_dict())
        return dict(edges=[dict(node=node) for node in nodes])

    def createAuthorization(
            self, name: str, description: str,
            organizationId: str, matching: str, inheritance: bool,
    ) -> dict:
        organization = self._organizations[organizationId]
        authorization = Authorization(
            id=gen_uuid(), name=name, description=description,
            organization=organization, inheritance=inheritance, matching=Matching(matching),
        )
        self._authorizations[authorization.id] = authorization
        self._authorization_permissions[authorization.id] = set()
        return dict(authorization=authorization.to_dict())

    def updateAuthorization(
            self, authorizationId: str,
            addPermissionId: str = None, removePermissionId: str = None,
    ):
        if addPermissionId is not None:
            self._authorization_permissions[authorizationId].add(addPermissionId)
        if removePermissionId is not None:
            self._authorization_permissions[authorizationId].discard(removePermissionId)
        return dict(authorization=self._authorizations[authorizationId].to_dict())

    def deleteAuthorization(self, authorizationId: str) -> dict:
        if authorizationId in self._authorizations:
            del self._authorizations[authorizationId]
            del self._authorization_permissions[authorizationId]
            for role_id, authorization_id_set in self._role_authorizations.items():
                authorization_id_set.discard(authorizationId)
            for member_id, authorization_id_set in self._member_authorizations.items():
                authorization_id_set.discard(authorizationId)
            return dict(deleted=True)
        return dict(deleted=False)

    def role(self, id: str) -> dict:
        if id in self._roles:
            return self._roles[id].to_dict()
        raise QottoAuthApiError(f"Role {id} does not exists")

    def roles(
            self, name_Icontains: str = None,
            organization_Id: str = None,
            member_Id: str = None, authorization_Id: str = None,
    ) -> dict:
        nodes = []
        for role in self._roles.values():
            if name_Icontains is not None and name_Icontains.lower() not in role.name.lower():
                continue
            if organization_Id is not None and role.organization.id != organization_Id:
                continue
            if member_Id is not None and role.id not in self._member_roles[member_Id]:
                continue
            if authorization_Id is not None and authorization_Id not in self._role_authorizations[role.id]:
                continue
            nodes.append(role.to_dict())
        return dict(edges=[dict(node=node) for node in nodes])

    def createRole(self, name: str, description: str, organizationId: str, inheritance: bool) -> dict:
        organization = self._organizations[organizationId]
        role = Role(
            id=gen_uuid(), name=name, description=description, organization=organization, inheritance=inheritance
        )
        self._roles[role.id] = role
        self._role_authorizations[role.id] = set()
        return dict(role=role.to_dict())

    def updateRole(
            self, roleId: str,
            addAuthorizationId: str = None, removeAuthorizationId: str = None,
    ) -> dict:
        if addAuthorizationId is not None:
            self._role_authorizations[roleId].add(addAuthorizationId)
        if removeAuthorizationId is not None:
            self._role_authorizations[roleId].discard(removeAuthorizationId)
        return dict(role=self._roles[roleId].to_dict())

    def deleteRole(self, roleId: str) -> dict:
        if roleId in self._roles:
            del self._roles[roleId]
            del self._role_authorizations[roleId]
            for member_id, role_id_set in self._member_roles.items():
                role_id_set.discard(roleId)
            return dict(deleted=True)
        return dict(deleted=False)

    def member(self, id: str) -> dict:
        if id in self._members:
            return self._members[id].to_dict()
        raise QottoAuthApiError(f"Member {id} does not exists")

    def members(self, user_Id: str = None, organization_Id: str = None) -> dict:
        nodes = []
        for member in self._members.values():
            if user_Id is None or user_Id == member.user.id:
                if organization_Id is None or organization_Id == member.organization.id:
                    nodes.append(member.to_dict())
        return dict(edges=[dict(node=node) for node in nodes])

    def createMember(self, userId: str, organizationId: str, uuid: str = None, **kwargs) -> dict:
        uuid = uuid or gen_uuid()
        user = self._users[userId]
        organization = self._organizations[organizationId]
        member = Member(uuid, user, organization)
        self._members[member.id] = member
        self._member_authorizations[member.id] = set()
        self._member_roles[member.id] = set()
        return dict(member=member.to_dict())

    def updateMember(
            self, memberId: str,
            addAuthorizationId: str = None, removeAuthorizationId: str = None,
            addRoleId: str = None, removeRoleId: str = None,
    ):
        if addAuthorizationId is not None:
            self._member_authorizations[memberId].add(addAuthorizationId)
        if removeAuthorizationId is not None:
            self._member_authorizations[memberId].discard(removeAuthorizationId)
        if addRoleId is not None:
            self._member_roles[memberId].add(addRoleId)
        if removeRoleId is not None:
            self._member_roles[memberId].discard(removeRoleId)
        return dict(member=self._members[memberId].to_dict())

    def deleteMember(self, memberId: str) -> dict:
        if memberId in self._members:
            del self._members[memberId]
            del self._member_authorizations[memberId]
            del self._member_roles[memberId]
            return dict(deleted=True)
        return dict(deleted=False)

    def account(self, id: str) -> dict:
        if id in self._accounts:
            return self._accounts[id].to_dict()
        raise QottoAuthApiError(f"Account {id} does not exists")

    def accounts(
            self,
            user_Id: str = None, application_Id: str = None,
            enabled: bool = None,
    ) -> dict:
        nodes = []
        for account in self._accounts.values():
            if user_Id is not None and user_Id != account.user.id:
                continue
            if application_Id is not None and application_Id != account.application.id:
                continue
            if enabled is not None and enabled != account.enabled:
                continue
            nodes.append(account.to_dict())
        return dict(edges=[dict(node=node) for node in nodes])

    def createAccount(self, userId: str, applicationId: str) -> dict:
        user = self._users[userId]
        application = self._applications[applicationId]
        account = Account(id=gen_uuid(), user=user, application=application, enabled=True, data={})
        self._accounts[account.id] = account
        return dict(account=account.to_dict())

    def updateAccount(self, accountId: str, setEnabled: bool = None) -> dict:
        account = self._accounts[accountId]
        if setEnabled is not None and setEnabled != account.enabled:
            account = account.update(enabled=setEnabled)
            self._accounts[accountId] = account
        return dict(account=account.to_dict())

    def deleteAccount(self, accountId: str) -> dict:
        if accountId in self._accounts:
            del self._accounts[accountId]
            return dict(deleted=True)
        return dict(deleted=False)

    def createUserRequest(self, identityId: str, comment: str = None, **kwargs) -> dict:
        identity = self._identities[identityId]
        if self.userRequests(identityId=identityId)['edges']:
            raise QottoAuthApiError(f"User request for identity {identityId} already exists")
        request = UserRequest(id=gen_uuid(), name=kwargs.pop('name', identity.name), comment=comment)
        self._user_requests[request.id] = request
        self._identities[identityId] = identity.update(user_request=request)
        return dict(userRequest=request.to_dict())

    def userRequest(self, id: str) -> dict:
        if id in self._user_requests:
            return self._user_requests[id].to_dict()
        raise QottoAuthApiError(f"User request {id} does not exists")

    def userRequests(self, identityId: str = None) -> dict:
        nodes = []
        for request in self._user_requests.values():
            if identityId is not None:
                for identity in self._identities.values():
                    if identityId == identity.id:
                        keep = bool(identity.user_request and identity.user_request.id == request.id)
                        break
                else:
                    keep = False
                if not keep:
                    continue
            nodes.append(request.to_dict())
        return dict(edges=[dict(node=node) for node in nodes])

    def deleteUserRequest(self, requestId: str) -> dict:
        if requestId in self._user_requests:
            del self._user_requests[requestId]
            return dict(deleted=True)
        return dict(deleted=False)

    def acceptUserRequest(self, userRequestId: str) -> dict:
        request = self._user_requests[userRequestId]
        del self._user_requests[userRequestId]
        for i in self._identities.values():
            if i.user_request and i.user_request.id == request.id:
                identity = i
                break
        else:
            raise QottoAuthApiError(f"Identity for user request {userRequestId} does not exists")
        user_data = self.createUser(name=request.name, identityId=identity.id)
        user = self._users[user_data['user']['id']]
        self.updateUser(user.id, addIdentityId=identity.id)
        identity = self._identities[identity.id]
        return dict(user=user.to_dict(), identity=identity.to_dict())

    def rejectUserRequest(
            self, userRequestId: str,
            blockIdentity: bool = False,
            deleteIdentity: bool = False,
    ) -> dict:
        request = self._user_requests[userRequestId]
        for i in self._identities.values():
            if i.user_request and i.user_request.id == request.id:
                identity = i
                break
        else:
            raise QottoAuthApiError(f"Identity for user request {userRequestId} does not exists")
        del self._user_requests[userRequestId]
        if blockIdentity:
            self.updateIdentity(identity.id, setBlocked=True)
        if deleteIdentity:
            self.deleteIdentity(identity.id)

        opt_identity = self._identities.get(identity.id)
        return dict(
            identity=opt_identity.to_dict() if opt_identity else None,
        )

    def identityTicket(self, identityId: str) -> str:
        if identityId in self._identity_tickets:
            return self._identity_tickets[identityId]
        raise QottoAuthApiError(f"Identity ticket {identityId} does not exists")

    def checkIdentityTicket(self, identityTicket: str) -> str:
        for identityId, ticket in self._identity_tickets.items():
            if ticket == identityTicket:
                return identityId
        raise QottoAuthApiError(f"Identity ticket {identityTicket} does not exists")

    def actor(self, tokenCookie: str = None, secretCookie: str = None) -> dict:
        if not tokenCookie or not secretCookie:
            return Actor().to_dict()
        payload = jwt.decode(tokenCookie, self._token_public_key, algorithms=[self._token_algorithm])
        computed_secret = jwt.encode(
            dict(token=tokenCookie), self._secret_private_key, algorithm=self._secret_algorithm
        )
        if computed_secret != secretCookie:
            return Actor().to_dict()

        user = self._users.get(payload.get('sub', ''))
        if user is None:
            return Actor().to_dict()

        member = self._members.get(payload.get('metadata', {}).get('member_id', ''))
        if member is None:
            return Actor(user=user).to_dict()

        return Actor(user=user, member=member).to_dict()

    def cookies(self, userId: str = None, organizationId: str = None) -> dict:
        user: Optional[User] = None
        member: Optional[Member] = None
        if userId and organizationId:
            for member in self._members.values():
                if member.organization.id == organizationId and member.user.id == userId:
                    user = member.user
                    member = member
                    break
            else:
                raise QottoAuthApiError(f"Member {userId} does not exists in organization {organizationId}")
        elif userId:
            if userId in self._users:
                user = self._users[userId]
            else:
                raise QottoAuthApiError(f"User {userId} does not exists")

        if not user:
            return CookiePair(
                token_cookie=Cookie(
                    name='token', value='', domain='localhost',
                    max_age=0, secure=False, http_only=False,
                ),
                secret_cookie=Cookie(
                    name='secret', value='', domain='localhost',
                    max_age=0, secure=False, http_only=True,
                ),
            ).to_dict()

        payload: dict[str, Any] = dict(
            iss='test',
            sub=user.id,
            uuid=user.id,
            is_superuser=user.is_superuser,
            metadata=dict(
                displayed_name=user.name,
                given_name='given',
                middle_name='middle',
                family_name='family',
                email='email',
            ),
            applications=[
                account.application.name
                for account in self._accounts.values()
                if account.user.id == user.id and account.enabled
            ],
        )
        if member:
            payload['org'] = member.organization.id
            payload['metadata']['member_id'] = member.id
            payload['metadata']['organization_id'] = member.organization.id
            payload['metadata']['organization_name'] = member.organization.name
            payload['metadata']['organization_namespace'] = str(member.organization.namespace)

        token = jwt.encode(payload, self._token_private_key, algorithm=self._token_algorithm)
        secret = jwt.encode(dict(token=token), self._secret_private_key, algorithm=self._token_algorithm)

        return CookiePair(
            token_cookie=Cookie(
                name='token', value=token, domain='localhost',
                max_age=0, secure=False, http_only=False,
            ),
            secret_cookie=Cookie(
                name='secret', value=secret, domain='localhost',
                max_age=0, secure=False, http_only=True,
            ),
        ).to_dict()

    def isAuthorized(
            self,
            actorId: str,
            permissionId: str,
            organizationId: str = None,
    ) -> bool:
        user: User
        if actorId in self._members:
            user = self._members[actorId].user
        elif actorId in self._users:
            user = self._users[actorId]
        else:
            return False
        if user.is_superuser:
            return True
        permission = self._permissions[permissionId]
        namespace: Optional[Namespace] = None
        if organizationId in self._organizations:
            namespace = self._organizations[organizationId].namespace
        elif organizationId is not None:
            namespace = Namespace(organizationId)

        for member in self._members.values():
            if not member.user.id == user.id:
                continue
            all_authorizations: set[Authorization] = set()
            member_namespace = member.organization.namespace
            for authorization_id in self._member_authorizations[member.id]:
                all_authorizations.add(self._authorizations[authorization_id])
            for role_id in self._member_roles[member.id]:
                role = self._roles[role_id]
                for authorization_id in self._role_authorizations[role.id]:
                    all_authorizations.add(self._authorizations[authorization_id])
            for authorization in all_authorizations:
                authorization_matching = authorization.matching
                if permission.id in self._authorization_permissions[authorization.id]:
                    if not namespace:
                        return True
                    if namespace.matches(member_namespace, authorization_matching):
                        return True
        return False

    def organizationsWhereIsAuthorized(
            self,
            actorId: str,
            permissionId: str,
    ) -> list:
        return [
            organization.to_dict()
            for organization in self._organizations.values()
            if self.isAuthorized(actorId, permissionId, organization.id)
        ]
